// Ethernet (Wiznet W5500) specific routines for USB+Ethernet/DMX Generator  **
//
// Simple DMX512(incl. Art-Net) controller for STM32 Nucleo-L476 + PCB
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the essential features of Art-Net, the MiniDMX protocol
// plus simple MIDI message and ASCII text based command sets
//
// Author and (c)2017-18 Wolfgang Schemmert. The copyright claim and GPL-3 Licence
// is restricted to this file and the depending files
// "main.c", "main.h"  "usb-otg.c", "dmx.c",and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics, Atollic AB and EMBITZ.org
//
// Contact: www.midi-and-more.de
// A detailled hardware construction and operation manual
// is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// **************************************************************************

#include "stm32l4xx.h"
#include "globaldefines.h"

//main
extern struct PARAM{
uint32_t VidPid;
uint32_t MacHi;
uint32_t MacLo;
uint32_t LocalIP;
uint32_t Subnet;
uint32_t Gateway;
uint32_t ArtnetIP;
uint32_t ArtSubnet;
uint32_t ArtUniverse;
uint32_t Channel;
uint32_t Reserved;
} gComParam;  //10 words

extern uint16_t gGREEN_LEDCOUNT, gTIMEOUT;
//extern void wait1ms(void);
extern void wait1us(void);
extern void servSystem(void);
//USART
extern void msDelay(uint16_t) ;
void sendCmdB(uint8_t);
extern void sendX8(uint8_t);
extern void sendX16(uint16_t);
extern void sendX32(uint32_t);

//dmx.c
extern uint8_t gdDMXBUF[DMX_BUFLEN+7];      //actual DMX transmit buffer
extern uint8_t gdFINALMEM[DMX_BUFLEN+7];    //final level of fade process
// ************************************

uint8_t geARTDMXSTR[14] = {'A','r','t','-','N','e','t',0,0,0x50,0,14,0,0} ;
uint8_t geSHORTNAME[18] = {'U','S','B','/','D','M','X','+','A','r','t',' ','G','e','n',0,0,0};
uint8_t geLONGNAME[24] = {'U','S','B','/','D','M','X','+','A','r','t','n','e','t',' ',
                          'G','e','n','e','r','a','t','o','r'};
uint8_t geREPORTSTR[13] = {'#','0','0','0','1','[','0','0','0','0',']','o','k'};

/*
.DB "Art-Net",0,0,0x50,0,14  ;12 Bytes
		;Byte 0..7 = Artnet ID
		;Byte 8 = Opcode Low
		;byte 9 = Opcode High ("OpOutput" = 0x5000)
		;Byte10 = ProtVerH = 0
		;Byte11 = ProtVerL = 14
		;Byte12 = Sequence 0=disabled
		;Byte13 = Physical (I/O num.)
*/
#define SPI_BUFLEN 1536
//SPI buffers are linear, not cyclic!!
uint8_t geSPI_RXBUF[SPI_BUFLEN];
uint8_t *geSPI_RXHEAD=0, *geSPI_RXTAIL=0, *geSPI_RXBUFTOP= &(geSPI_RXBUF[SPI_BUFLEN]);
uint8_t geSPI_TXBUF[SPI_BUFLEN];

uint8_t geDummy;            //global only to avoid warning during compilation
uint8_t geArtnetMsg;        //special Art-Net message types
uint8_t geArtnetOn;         //1=default: Artnet data are send to DMX. 0=block

uint8_t initW5500(void);
uint8_t openUDP0(void);
uint8_t addEtherQuery(uint8_t);
uint8_t checkEtherSend(void);
uint8_t sendEtherQuery(void);
uint8_t sendEtherByte(uint8_t);
uint16_t sendEtherBuffer(uint8_t *, uint16_t);
uint8_t receiveUDP(void);
void artPollReply(void);
void handleArtAddress(void);

void initSPI1(void);
void sendSpiWord(uint16_t);
uint16_t receiveSpiWord(void);
uint16_t exchangeSpiWord(uint16_t);
uint8_t exchangeSpiByte(uint8_t);
void writeWiz_4(uint16_t, uint8_t, uint8_t);
uint8_t readWiz_4(uint16_t, uint8_t);
void writeWiz_5(uint16_t, uint8_t, uint16_t);
uint16_t readWiz_5(uint16_t, uint8_t);
void writeWiz_7(uint16_t, uint8_t, uint32_t);
uint32_t readWiz_7(uint16_t, uint8_t);
uint16_t writeWiz_Data(uint16_t, uint8_t*, uint16_t);
uint16_t readWiz_Data(uint16_t, uint16_t);
uint16_t pullSpiRxBuf(void);

/*
//W5500 Common Registers (SPI Adressing)
//--------------------------------------
#define COMMON_BLOCK    0x00
#define MR              0x0000  //Mode Register, for S/W Reset
#define GAR0            0x0001  //4 Gateway Address Register
#define SUBR0           0x0005  //4 Subnet Mask Register
#define SHAR0           0x0009  //6 Source Hardware Address Register
#define SIPR0           0x000F  //4 Source IP Address Register
#define IR              0x0015  //Interrupt Register
#define IMR             0x0016  //Interrupt Mask Register
#define SIR             0x0017  //Socket Interrupt Register
#define SIMR            0x0018  //Socket Interrupt Mask Register
#define RTR0            0x0019  //Word Initial Retry Time-value Register
#define RCR             0x001B  //Retry Count Register
#define PHYCFGR         0x002E  //PHY Configuration Register

//W5500 Socket0 Registers
//-----------------------
#define S0_REGBLOCK     0x01
#define S0_MR           0x0000  //Socket 0 Mode Register
#define S0_CR           0x0001  //Socket 0 Command Register
#define S0_IR           0x0002  //Socket 0 Interrupt Register
#define S0_SR           0x0003  //Socket 0 Status Register
#define S0_PORT0        0x0004  //Socket 0 Local Port Register
#define S0_DHAR0        0x0006  //Socket 0 Destination Hardware Address
#define S0_DIPR0        0x000C  //Socket 0 Destination IP Address
#define S0_DPORT0       0x0010  //Socket 0 Destination Port Register
#define S0_MSSR0        0x0012  //Socket 0 Maximum Segment Size
#define S0_TTL			0x0016  //Socket 0 Time to Live Register
#define S0_RXBUF_SIZE   0x001E  //Socket 0 Receive Buffer Size
#define S0_TXBUF_SIZE   0x001F  //Socket 0 Transmit Buffer Size
#define S0_TX_FSR0      0x0020  //Socket 0 TX Free Size
#define S0_TX_RD0       0x0022  //Socket 0 TX Read Pointer
#define S0_TX_WR0       0x0024  //Socket 0 TX Write Pointer
#define S0_RX_RSR0      0x0026  //Socket 0 RX Received Size
#define S0_RX_RD0       0x0028  //Socket 0 RX Read Pointer
#define S0_RX_WR0       0x002A  //Socket 0 RX Write Pointer
#define S0_IMR          0x002C  //Socket 0 Interrupt Mask Register
//#define S0_KPALVTR      0x002F  //Socket 0 Keep Alive Timer (TCP only)

#define S0_TXBLOCK      0x02
#define S0_RXBLOCK      0x03
*/

//Art-Net specials
//----------------
/*
ARTDMXSTR:      .DB "Art-Net",0,0,0x50,0,14  ;12 Bytes
		;Byte 0..7 = Artnet ID
		;Byte 8 = Opcode Low
		;byte 9 = Opcode High ("OpOutput")
		;Byte10 = ProtVerH = 0
		;Byte11 = ProtVerL = 14
		;Byte12 = Sequence 0=disabled
		;Byte13 = Physical (I/O num.)

;ber ArtADDR modifizierbar:
SHORTNAME: .DB "MIDI-RS-E/DMX Gen",0    ;18 incl Nullbyte
LONGNAME:  .DB "MIDI-RS-Ethernet/DMX Generator",0,0,0,0

	   .DW 0,0,0,0,0,0,0,0

	   .DW 0,0,0,0,0,0,0,0
REPORTSTR: .DB "#0001[0000]ok",0,0,0          ;16 incl Nullbyte
	   .DW 0,0,0,0,0,0,0,0
	   .DW 0,0,0,0,0,0,0,0
	   .DW 0,0,0,0,0,0,0,0

*/

// ########################################################################

uint8_t initW5500(void)   // **********************************************
// initializes Wiz5500 registers for desired mode of operation.
// **************************************************************************
{
uint32_t artSubnetMask;

//PortC,4 handles W5500 /RST
    GPIOC->BSRR = GPIO_BSRR_BR4;                               //set /RST low
    msDelay(5);
    GPIOC->BSRR = GPIO_BSRR_BS4;                               //set /RST high
    msDelay(100);                                               //min 50ms

    if (gComParam.ArtnetIP == 0) artSubnetMask = gComParam.Subnet; //overwrite
    else artSubnetMask = 0xFF000000;
    msDelay(100);
//allocate maximum socket memory (16kB)
    writeWiz_4(S0_RXBUF_SIZE,S0_REGBLOCK,16);
    writeWiz_4(S0_TXBUF_SIZE,S0_REGBLOCK,16);
//transfer MAC ID from gComParam
    writeWiz_5(SHAR0,COMMON_BLOCK,(uint16_t)(gComParam.MacHi & 0x0000FFFF));
    writeWiz_7(SHAR0+2,COMMON_BLOCK,gComParam.MacLo);
//transfer Gateway IP from gComParam
    writeWiz_7(GAR0,COMMON_BLOCK,gComParam.Gateway);
//transfer Subnet Mask from gComParam
    writeWiz_7(SUBR0,COMMON_BLOCK,artSubnetMask);
//transfer Local IP from gComParam
    writeWiz_7(SIPR0,COMMON_BLOCK,gComParam.LocalIP);

    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);             //clear pending interrupts
    return 1;
}


uint8_t openUDP0(void)   // *************************************************
// with entries from gComParam
// **************************************************************************
{
uint8_t recvbyte;
uint32_t artDestIP, artIpB;

    if(gComParam.ArtnetIP==0) artDestIP=gComParam.LocalIP | ~(gComParam.Subnet);
    else  {
        artIpB = (((gComParam.MacLo>>16) + OEMH + OEML) & 0xFF) << 16;
        sendX32(artIpB);
        artDestIP =(gComParam.ArtnetIP<<24)|artIpB|(gComParam.MacLo&0x0000FFFF);
    }

//transfer UDP Local Port from gComParam
   writeWiz_5(S0_PORT0,S0_REGBLOCK,ARTNETPORT);
//transfer UDP Destination Port from gComParam
    writeWiz_5(S0_DPORT0,S0_REGBLOCK,ARTNETPORT);
//transfer UDP Destination IP from gComParam
    writeWiz_7(S0_DIPR0,S0_REGBLOCK,artDestIP);
//activate UDP mode
//UDP Unicast or Broadcast
    writeWiz_4(S0_MR,S0_REGBLOCK,0x02);             //normal UDP mode
    writeWiz_4(S0_CR,S0_REGBLOCK,0x01);                         //=CMD_OPEN
    recvbyte = 0;
    while ((recvbyte= readWiz_4(S0_SR,S0_REGBLOCK)) != 0x22 );  //=SOCK_UDP

    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);             //clear interrupt flags
    writeWiz_4(S0_IMR,S0_REGBLOCK,0);               //no interrupts
    writeWiz_4(SIR,COMMON_BLOCK,0);
    writeWiz_4(SIMR,COMMON_BLOCK,0);

    return 1;
}


uint8_t sendEtherByte(uint8_t data)  // *************************************
// **************************************************************************
{
uint16_t temp, txPtr;

    temp = readWiz_5(S0_TX_FSR0,S0_REGBLOCK);
    if (temp == 0) return 0;                        //no buffer available

    txPtr = readWiz_5(S0_TX_WR0,S0_REGBLOCK);
    writeWiz_4(txPtr,S0_TXBLOCK,data);
    txPtr++;
    writeWiz_5(S0_TX_WR0,S0_REGBLOCK,txPtr);        //update W5500 write pointer
    writeWiz_4(S0_CR,S0_REGBLOCK,0x20);             //=CMD_SEND
    writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);

    return 1;
}


uint16_t sendEtherBuffer(uint8_t *buffer, uint16_t toSend)  // **************
// demands linear buffer, not cyclic !
// **************************************************************************
{
uint16_t temp, txPtr;

    temp = readWiz_4(S0_SR,S0_REGBLOCK);
    temp = readWiz_5(S0_TX_FSR0,S0_REGBLOCK);
    if (temp < toSend) return 0;                        //no buffer available

    txPtr = readWiz_5(S0_TX_WR0,S0_REGBLOCK);
    txPtr = writeWiz_Data(txPtr, buffer, toSend);
    writeWiz_5(S0_TX_WR0,S0_REGBLOCK,txPtr);        //update W5500 write pointer
    writeWiz_4(S0_CR,S0_REGBLOCK,0x20);             //=CMD_SEND

    return 1;
}


uint8_t receiveUDP(void)  // ************************************************
//if available, one UDP packet is transferred
//from W5500 internal receive buffer to geSPI_RXBUF (linear,not cyclic!!)
//buffer pointers are updated
//Art-Net special information is evaluated
// **************************************************************************
{

uint16_t rxCount, rxPtr, udpLength, k;
uint32_t udpHeader;

    rxCount = readWiz_5(S0_RX_RSR0,S0_REGBLOCK);
    if (rxCount == 0) return 1;
    if (rxCount <= 8) return 0;                     //empty UDP packet or error

    rxPtr = readWiz_5(S0_RX_RD0,S0_REGBLOCK);
    udpHeader = readWiz_7(rxPtr,S0_RXBLOCK);
    rxPtr += 4;
    udpHeader = readWiz_7(rxPtr,S0_RXBLOCK);
    rxCount -= 8;
    rxPtr += 4;
    udpLength =(uint16_t)(udpHeader & 0x0000FFFF);
                                        //what if geUdpLength != rxPtr ????????
    geSPI_RXHEAD = geSPI_RXBUF;
    geSPI_RXTAIL = geSPI_RXBUF;
    rxPtr = readWiz_Data(rxPtr,udpLength);      //copy message into geSPI_RXBUF
    writeWiz_5(S0_RX_RD0,S0_REGBLOCK,rxPtr);    //update W5500 read pointer
    writeWiz_4(S0_CR,S0_REGBLOCK,0x40);         //=CMD_RECV

//process Art-Net message
    geArtnetMsg = geSPI_RXBUF[9];
    if (geArtnetMsg == 0x20) {
        artPollReply();
        return 1;
    }
    else if (geArtnetMsg != 0x50) return 0;
//handle Art-Net data packet:
//redefine udpLength
    udpLength = ((uint16_t)geSPI_RXBUF[16] << 8) + (uint16_t)geSPI_RXBUF[17];
    if (geArtnetOn)  {                  //0 = block Artnet, only USB control
        for (k=0;k<udpLength;k++)  {
            gdDMXBUF[k+1] = geSPI_RXBUF[k+18];
			gdFINALMEM[k+1] = geSPI_RXBUF[k+18];
        }
        GPIOC->BSRR = GPIO_BSRR_BR_2; // clear bit PC2, green LED OFF
        gGREEN_LEDCOUNT = 200; //milliseconds
    }
    return 1;
}


void artPollReply(void)   // ************************************************
// **************************************************************************
{
uint16_t k;
uint8_t n;
uint32_t temp;

    k = 0;
    n = 0;
    for (n=0;n<9;n++) {
        geSPI_TXBUF[k++] = geARTDMXSTR[n];
    }
    geSPI_TXBUF[k++] = 0x21;        //Opcode "ArtPollReply"
    temp = gComParam.ArtnetIP;
    n = (temp>>24) & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = (temp>>16) & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = (temp>>8) & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = temp & 0x000000FF;
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)(ARTNETPORT & 0x00FF);
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((ARTNETPORT >> 8) & 0x00FF);
    geSPI_TXBUF[k++] = n;
    geSPI_TXBUF[k++] = HARDWARE_VERSION;    //byte 16
    geSPI_TXBUF[k++] = REVNUM;
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = (uint8_t)(gComParam.ArtSubnet & 0x0000000F);
    geSPI_TXBUF[k++] = OEMH;
    geSPI_TXBUF[k++] = OEML;
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = 0xD0;    //Node Status= Indicators normal
                                //Front Panel setup, kein Dual Boot
    geSPI_TXBUF[k++] = 0x69;    //'i' Cinetix ESTA Low  byte 24
    geSPI_TXBUF[k++] = 0x43;    //'C' Cinetix ESTA High
//REMOTESHORT einplanen
    for (n=0;n<18;n++) {
        geSPI_TXBUF[k++] = geSHORTNAME[n];  //bytes 26 to 43
    }
//REMOTELONG einplanen
    for (n=0;n<24;n++) {
        geSPI_TXBUF[k++] = geLONGNAME[n];   //bytes 44-107 (total 64char)
    }
        for (n=0;n<40;n++) {
        geSPI_TXBUF[k++] = 0;               //filling bytes of LONGNAME
    }
    for (n=0;n<13;n++) {
        geSPI_TXBUF[k++] = geREPORTSTR[n];   //bytes 108-171 (total 64char)
    }
        for (n=0;n<51;n++) {
        geSPI_TXBUF[k++] = 0;               //filling bytes of REPORTSTR
    }
    geSPI_TXBUF[k++] = 0;       //NumPortsH = byte 172
    geSPI_TXBUF[k++] = 1;       //NumPortsL = byte 173

    geSPI_TXBUF[k++] = 0x80;    //PortTypes: one DMX512 output byte 174
    geSPI_TXBUF[k++] = 0;       //fills bytes 175-177
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = 0;

    geSPI_TXBUF[k++] = 0;       //Good Input =0 bytes 178-181
    geSPI_TXBUF[k++] = 0;       //fills
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = 0;

    geSPI_TXBUF[k++] = 0x80;    //Good Output  bytes 182-185
    geSPI_TXBUF[k++] = 0;       //fills
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = 0;

    geSPI_TXBUF[k++] = 0;       //Swin: dummy:no Rx Uinverse bytes 186-189
    geSPI_TXBUF[k++] = 0;       //fills
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = 0;

    n= (gComParam.ArtSubnet & 0x0000000F)<<4;
    n |= gComParam.ArtUniverse & 0x0000000F;
    geSPI_TXBUF[k++] = n;       //Swout  bytes 190 - 193
    geSPI_TXBUF[k++] = 0;       //fills
    geSPI_TXBUF[k++] = 0;
    geSPI_TXBUF[k++] = 0;

    for (n=0;n<6;n++) {
        geSPI_TXBUF[k++] = 0;   //191-199: div. special features = 0
    }
    geSPI_TXBUF[k++] = 0;       //Style: "a DMX to / from Art-Net device"
    n = (uint8_t)((gComParam.MacHi >> 8) & 0x000000FF);  //201
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)(gComParam.MacHi& 0x000000FF);  //202
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((gComParam.MacLo >> 24) & 0x000000FF);  //203
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((gComParam.MacLo >> 16) & 0x000000FF);  //204
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)((gComParam.MacLo >> 8) & 0x000000FF);  //205
    geSPI_TXBUF[k++] = n;
    n = (uint8_t)(gComParam.MacLo & 0x000000FF);  //206
    geSPI_TXBUF[k++] = n;
    for (n=0;n<34;n++) {
        geSPI_TXBUF[k++] = 0;   //207-240
    }
    sendEtherBuffer(geSPI_TXBUF,238);

    GPIOC->BSRR = GPIO_BSRR_BR_2; // clear bit PC2, green LED OFF
    gGREEN_LEDCOUNT = 200; //milliseconds
}


// ##########################################################################
// low level routines

void initSPI1(void)   // ****************************************************
//SPI1 is Master
// **************************************************************************
{

     RCC->APB2ENR |=  RCC_APB2ENR_SPI1EN ;   // SPI1 clock enable

//PB3 is SPI1_SCK
//PB4 is SPI1_MISO
//PB5 is SPI1-MOSI, configure as Alternate Function
    GPIOB->MODER &= ~(GPIO_MODER_MODE3 | GPIO_MODER_MODE4 | GPIO_MODER_MODE5);
    GPIOB->MODER |= (GPIO_MODER_MODE3_1|GPIO_MODER_MODE4_1|GPIO_MODER_MODE5_1);

//PB13 is /INT input (not used for this application)
//PB14 is /SCN output
    GPIOB->MODER &= ~(GPIO_MODER_MODE13 | GPIO_MODER_MODE14);
    GPIOB->MODER |= (GPIO_MODER_MODE14_0);
    GPIOB->OTYPER &= ~(GPIO_OTYPER_OT14);   //PB14 push-pull
    GPIOB->PUPDR &= ~GPIO_PUPDR_PUPDR13;
    GPIOB->PUPDR |= GPIO_PUPDR_PUPDR13_0;   //activate pull-up
//set GPIOB 3,4,5,13,14 Speed up to 25MHz ("middle speed")
    GPIOB->OSPEEDR &= 0xC3FFF03F ;          //OSPEEDRx_1 down
    GPIOB->OSPEEDR |= 0x14000540 ;          //OSPEEDRx_0 up
//PC4 is /RST output
    GPIOC->MODER &= ~(GPIO_MODER_MODE4);
    GPIOC->MODER |= (GPIO_MODER_MODE4_0);
    GPIOC->OTYPER &= ~(GPIO_OTYPER_OT4);    //PC4 push-pull
//set GPIOC 4 Speed up to 25MHz ("middle speed")
    GPIOC->OSPEEDR &= ~GPIO_OSPEEDR_OSPEED4;   //OSPEEDR4 down
    GPIOC->OSPEEDR |= GPIO_OSPEEDR_OSPEED4_0;  //OSPEEDR4_0 up
    GPIOC->BSRR = GPIO_BSRR_BS_4;               //set /RST high

//SPI1 is AF5
    GPIOB->AFR[0] &= 0xFF000FFF;            //configure PB3,4,5 AF function
    GPIOB->AFR[0] |= 0x00555000;

    geSPI_RXHEAD = geSPI_RXBUF;
    geSPI_RXTAIL = geSPI_RXBUF;

//SPI Master configuration
    SPI1->CR1 = SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_MSTR;
                                //idle=low,sample rising edge,freq=1/2APB2 clock
    SPI1->CR2 =  SPI_CR2_DS;    //default 16bit, polling

    GPIOC->BSRR = GPIO_BSRR_BS4;                                //set /RST high
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high

    SPI1->CR1 |= SPI_CR1_SPE;                                   //enable
}

/*
//was contained in previous firmware versions and worked acceptably
//removed, because critical and somewhat obscure together with STM32

uint8_t exchangeSpiByte(uint8_t data)   // **********************************
// **************************************************************************
{
uint8_t recv;

    SPI1->CR2 = SPI_CR2_FRXTH | 0x0700 ;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    while ((SPI1->SR & SPI_SR_TXE) == 0);
    SPI1->DR = (uint16_t)data;
    while ((SPI1->SR & SPI_SR_RXNE) == 0);
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
//STM32 processors support 8 bit SPI transfer very badly, always 16 clock pulses
//for this reason, when data blocks with even count are transferred,
//exchangeSpiByte() must always be the last transfer command
//and /SCN must be set after 8bits are transferred
//if possible, avoid exchangeSpiByte()
    recv = SPI1->DR;
    wait1us();
//???????? special hack for fine adjustment of pulse edges ?????
//seemed necessary for a while during firmware development,
//but now it works without. Exact reason unknown.

    while (SPI1->SR & SPI_SR_RXNE) {
        geDummy = SPI1->DR;
    }

    SPI1->CR2 =  SPI_CR2_DS;
    return recv;
}
*/

uint16_t exchangeSpiWord(uint16_t data)   // ********************************
//this is well compatible with 16bit SPI clock burst
// **************************************************************************
{
    while ((SPI1->SR & SPI_SR_TXE) == 0);
    SPI1->DR = data;
    while ((SPI1->SR & SPI_SR_RXNE) == 0);
    return SPI1->DR;
}


void writeWiz_4(uint16_t address, uint8_t block, uint8_t data)   // *******
// write 4 byte Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    send = ((block<<11) | 0x0400 | data);
    exchangeSpiWord(address);
    exchangeSpiWord(send) ;

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high

}


void writeWiz_5(uint16_t address, uint8_t block, uint16_t data) // **********
// write 5 byte Wiz5500 compatible cluster (1 halfword, e.g. port no.)
// **************************************************************************
{
uint16_t send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
    send = (block<<11) | 0x0400 | (data >> 8) ;
    exchangeSpiWord(address);
    exchangeSpiWord(send) ;

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    send = (block<<11) | 0x0400 | (data & 0x00FF);
    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
                                                // /SCN is high for about 1 us
    exchangeSpiWord(address+1);
    exchangeSpiWord(send) ;

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
}


void writeWiz_7(uint16_t address, uint8_t block, uint32_t data)  // *********
// write 7 byte Wiz5500 compatible cluster (1 word, e.g. IP address)
// **************************************************************************
{
    writeWiz_5(address,block,(uint16_t)((data >> 16) & 0x0000FFFF));
    writeWiz_5(address+2,block,(uint16_t)(data & 0x0000FFFF));
}


uint8_t readWiz_4(uint16_t address, uint8_t block)   // *******************
// read 1 databyte Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t recv, send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    exchangeSpiWord(address);
    send = (block<<11);
    recv = exchangeSpiWord(send) ;
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high

    return ((uint8_t)(recv & 0x00FF));
}


uint16_t readWiz_5(uint16_t address, uint8_t block)   // ********************
// read 1 halfword Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t send, recv0, recv1;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    send = (block<<11);
    exchangeSpiWord(address);
    recv0 = exchangeSpiWord(send);

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    send = (block<<11);
    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
                                                // /SCN is high for about 0.5 us
    exchangeSpiWord(address+1);
    recv1 = exchangeSpiWord(send);

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high

    return ((recv0 & 0x00FF) << 8) | (recv1 & 0x00FF);
}


uint32_t readWiz_7(uint16_t address, uint8_t block)   // ********************
// read 1 word Wiz5500 compatible cluster
// **************************************************************************
{
uint32_t  recv2, recv0;
    recv0 = (uint32_t)readWiz_5(address,block);
    recv2 = (uint32_t)readWiz_5(address+2,block);
    return ((recv0 << 16) | (recv2 & 0x0000FFFF));
}


uint16_t writeWiz_Data(uint16_t address, uint8_t* buffer, uint16_t toSend) // *
// write (toSend+3) byte Wiz5500 compatible cluster (i.e. toSend databytes)
// **************************************************************************
{
uint16_t send;
uint8_t byte1, byte2;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    if ((toSend & 1) == 0)  {               //an even number of bytes to be sent
    send = ((S0_TXBLOCK<<11)|0x0400 |*buffer++); //linear buffer, no wrap around
    exchangeSpiWord(address);
    exchangeSpiWord(send);
    toSend--;
    address++;
    }

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    send = ((S0_TXBLOCK<<11) | 0x0400 | *buffer++);

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
                                                // /SCN is high for about 1 us
    exchangeSpiWord(address);
    exchangeSpiWord(send);
    toSend--;
    address++;

//from here, only zero or an even number of bytes to be sent
    while (toSend)  {
        byte1 = *buffer++;
        byte2 = *buffer++;
        exchangeSpiWord(byte1<<8 | byte2);
        toSend -= 2;
        address += 2;
    }
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    return address;
}


uint16_t readWiz_Data(uint16_t address,uint16_t toRead)  // *****************
// read (count) databytes from Wiz5500 Rx Buffer to geSPI_RXBUF
// **************************************************************************
{
uint16_t recvword, send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    if ((toRead & 1) == 0)   {              //an even number of bytes to be read
        send = (S0_RXBLOCK<<11);
        exchangeSpiWord(address);
        recvword = exchangeSpiWord(send);
        *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        toRead--;
        address++;
    }

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    send = (S0_RXBLOCK<<11);
    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
                                            // /SCN is high for about 0.5 us
    exchangeSpiWord(address);
    recvword = exchangeSpiWord(send);
    *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
    geSPI_RXHEAD++;
    if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
    toRead--;
    address++;

//from here, only zero or an even number of bytes to be read
    while (toRead)  {
        recvword = exchangeSpiWord(0);
        *geSPI_RXHEAD = (uint8_t)((recvword >> 8) & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);
        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        toRead -= 2;
        address += 2;
    }
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    return address;
}

/*
uint8_t readWiz_4(uint16_t address, uint8_t block)   // *******************
// read 1 databyte Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t recv, send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    exchangeSpiWord(address);
    send = (block<<11);
    recv = exchangeSpiWord(send) ;
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high

    return ((uint8_t)(recv & 0x00FF));
}


void writeWiz_5(uint16_t address, uint8_t block, uint16_t data)
// write 5 byte Wiz5500 compatible cluster (1 halfword, e.g. port no.)
// **************************************************************************
{
uint16_t send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    exchangeSpiWord(address);
    send = (block<<11) | 0x0400 | (data >> 8) ;
    exchangeSpiWord(send) ;
    exchangeSpiByte(data & 0x00FF);

  	GPIOB->BSRR = GPIO_BSRR_BS14;       						//set /SCN high
}


uint16_t readWiz_5(uint16_t address, uint8_t block)   // *******************
// read 1 halfword Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t recv1, send;
uint8_t  recv0;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low

    exchangeSpiWord(address);
    send = (block<<11);
    recv1 = exchangeSpiWord(send) ;
    recv0 = exchangeSpiByte(0);
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    return (((recv1 & 0x00FF)<<8) | recv0);
}


void writeWiz_7(uint16_t address, uint8_t block, uint32_t data)
// write 7 byte Wiz5500 compatible cluster (1 word, e.g. IP address)
// **************************************************************************
{
uint16_t send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
    exchangeSpiWord(address);
    send = (block<<11) | 0x0400 | (data >> 24) ;
    exchangeSpiWord(send);
    send = (data >> 8) & 0x0000FFFF;
    exchangeSpiWord(send) ;
    exchangeSpiByte(data & 0x000000FF);

    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
}


uint32_t readWiz_7(uint16_t address, uint8_t block)   // *******************
// read 1 word Wiz5500 compatible cluster
// **************************************************************************
{
uint16_t  send;
uint32_t  recv3, recv12, recv0;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
    exchangeSpiWord(address);
    send = (block<<11);
    recv3 = (uint32_t)exchangeSpiWord(send) ;
    recv12 = (uint32_t)exchangeSpiWord(0);
    recv0 = (uint32_t)exchangeSpiByte(0);
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    return (((recv3 & 0x000000FF)<<24) | ((recv12 & 0x0000FFFF)<<8) | recv0);
}


uint16_t writeWiz_Data(uint16_t address, uint8_t* buffer, uint16_t toSend)
// write (toSend+3) byte Wiz5500 compatible cluster (i.e. toSend databytes)
// **************************************************************************
{
uint16_t send;
uint8_t byte1, byte2;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
    exchangeSpiWord(address);
    send = ((S0_TXBLOCK<<11) | 0x0400 | *buffer++);
    exchangeSpiWord(send);
    toSend--;
    address++;
    while (toSend > 1)  {
        byte1 = *buffer++;
        byte2 = *buffer++;
        exchangeSpiWord(byte1<<8 | byte2);
        toSend -= 2;
        address += 2;
    }
    if (toSend==1) {
        exchangeSpiByte(*buffer);
        address++;
    }
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    return address;
}


uint16_t readWiz_Data(uint16_t address,uint16_t toRead)
// read (count) databytes from Wiz5500 Rx Buffer to geSPI_RXBUF
// **************************************************************************
{
uint16_t recvword, send;

    GPIOB->BSRR = GPIO_BSRR_BR14;                               //set /SCN low
    exchangeSpiWord(address);

    send = (S0_RXBLOCK<<11);
    recvword = exchangeSpiWord(send);
    *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);

    geSPI_RXHEAD++;
    if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
    toRead--;
    address++;

    while (toRead > 1)  {
        recvword = exchangeSpiWord(0);
        *geSPI_RXHEAD = (uint8_t)((recvword >> 8) & 0x00FF);

        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        *geSPI_RXHEAD = (uint8_t)(recvword & 0x00FF);

        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        toRead -= 2;
        address += 2;
    }
    if (toRead==1) {
        *geSPI_RXHEAD = exchangeSpiByte(0);

        geSPI_RXHEAD++;
        if (geSPI_RXHEAD >= geSPI_RXBUFTOP)  geSPI_RXHEAD = geSPI_RXBUF;
        address++;
    }
    GPIOB->BSRR = GPIO_BSRR_BS14;                               //set /SCN high
    return address;
}

uint16_t pullSpiRxBuf(void) // *********************************************
// pull 1 byte asynchronously from SPI1 reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint16_t recv;

    servSystem();   //serve watchdog etc.

    if (geSPI_RXTAIL != geSPI_RXHEAD)  {
        recv= (uint16_t)((*geSPI_RXTAIL++ & 0x00FF)|0x8000);
        if (geSPI_RXTAIL == geSPI_RXBUFTOP) geSPI_RXTAIL=geSPI_RXBUF;
        return recv;
    }
    else return 0;
}
*/

//###########################################################################

