// User interface specific routines for USB+Ethernet/DMX Generator  ***********
//
// Simple DMX512(incl. Art-Net) controller for STM32 Nucleo-L476 + PCB
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the essential features of Art-Net, the MiniDMX protocol
// plus simple MIDI message and ASCII text based command sets
//
// Author and (c)2017-18 Wolfgang Schemmert. The copyright claim and GPL-3 Licence
// is restricted to this file and the depending files
// "main.c", "main.h". "usb-otg.c", "dmx.c", and "ether.c".
// Other depending material is published and copyrighted
// by STMicroelectronics, Atollic AB and EMBITZ.org
//
// Contact: www.midi-and-more.de
// A detailled hardware construction and operation manual
// is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// **************************************************************************

#include "stm32L476xx.h"
#include "globaldefines.h"

// ************************************

extern struct PARAM{
uint32_t VidPid;
uint32_t MacHi;
uint32_t MacLo;
uint32_t LocalIP;
uint32_t Subnet;
uint32_t Gateway;
uint32_t ArtnetIP;
uint32_t ArtSubnet;
uint32_t ArtUniverse;
uint32_t Channel;
uint32_t Reserved;
} gComParam;  //10 words

extern uint16_t gVidShadow, gPidShadow;
extern void popFlashPage15(void);
extern void pushFlashPage15(void);
void popPresetPage(uint32_t);
void pushPresetPage(uint32_t);
extern void servSystem(void);
extern void msDelay(uint32_t);

//RS-232
extern void initUsart2(void);
extern void sendCmdB(uint8_t);
extern void sendX8(uint8_t);
extern void sendN8(uint8_t);
extern void sendX16(uint16_t);
extern void sendX32(uint32_t);
extern void sendS32(uint32_t);
extern void sendString(const uint8_t *);
extern uint16_t pullCmdRxBuf(void);




//USB
extern uint8_t guDeviceClass;

//ether.c
extern uint8_t geArtnetOn;  //1=default: Artnet data are send to DMX. 0=block

//dmx.c
extern uint16_t gdDmxAddr, gdDmxMax;
extern uint8_t gdDMXBUF[DMX_BUFLEN+7];          //actual DMX transmit buffer
extern uint8_t gdFINALMEM[DMX_BUFLEN+7];        //final level of fade process
extern uint16_t gdFadeTime;
extern void setDmxAddress(void);
extern void setDmxLevel(uint8_t);
extern void setNextDmxLevel(uint8_t);
extern void setEqual(uint8_t);
extern void msgDmxState(void);
extern void readDmxBuf(void);
extern void sendAddrLevel(void);
extern void sendLevel(uint16_t);

// ************************************

uint32_t gcEtherIP;
uint8_t gcPushByte, gcIgnoreVelo0;
const uint8_t gcSetStr[] = {"set"};
const uint8_t gcVidPidStr[] = {" USB Vid/Pid of this device (hex)"};
const uint8_t gcMacIdStr[] = {" Ethernet MAC ID (hex)m5.m4.m3.m2.m1.m0)"};
const uint8_t gcGatewayStr[] = {" Gateway IP (ip3.ip2.ip1.ip0) "};
const uint8_t gcLocalIpStr[] = {" IP of this device (ip3.ip2.ip1.ip0)"};
const uint8_t gcSubnetStr[] = {" Subnet Mask of this device (m3.m2.m1.m0)"};
const uint8_t gcArtIpStr[] = {" Art-Net 'IP Addr A' (0, 2, 10, default=0)"};
const uint8_t gcArtSubnetStr[] = {" Art-Net Subnet (0-15)"};
const uint8_t gcArtUnivStr[] = {" Art-Net Universe DMX OUT (0-15)"};
const uint8_t gcChannelStr[] = {" MIDI base channel (1-16)"};

void  AsciiCmd_Interpreter(uint8_t);
void  MidiCmd_Interpreter(uint32_t);
void  Setup_Interpreter(uint8_t);
uint32_t parseHexByte(void);
uint32_t parseHexNibble(void);
uint32_t parseDeciNum(uint32_t);
uint8_t parseDeciNibble(void);
void sendS3(uint32_t);
void sendS2(uint32_t);
uint8_t setVidPid(void);
uint8_t setMacID(void);
uint8_t setEtherIP(void);
void setMidiChannel(void);
void listCommands(void);
void listSetup(void);


void AsciiCmd_Interpreter(uint8_t CmdCode)  // *********************************
// **************************************************************************
{
uint32_t k;
uint16_t recvword;
const uint8_t AgainStr[] = {"all lights DARK?->press again "};

    if ((CmdCode > 0x60) && (CmdCode < 0x7B)) CmdCode -= 0x20; //->upper case

    sendCmdB(CmdCode);

    switch (CmdCode)  {
    case 'S':
        setDmxAddress();
        break;

    case 'V':
        if ((k = parseDeciNum(255)) > 255)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else setDmxLevel(k);
        break;

    case 0x2C:      //comma
         if ((k = parseDeciNum(255)) > 255)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else setNextDmxLevel(k);
        break;

    case 'Q':
        msgDmxState();
        break;

    case 'R':
        readDmxBuf();
        break;

    case 'T':
        if ((k = parseDeciNum(127)) > 127)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else gdFadeTime = k * 5;
        break;

    case '+':
        if (gdDMXBUF[gdDmxAddr] != 0xFF) {
            gdDMXBUF[gdDmxAddr]++ ;
            gdFINALMEM[gdDmxAddr]++ ;
        }
        break;

   case '-':
        if (gdDMXBUF[gdDmxAddr] != 0x00) {
            gdDMXBUF[gdDmxAddr]-- ;
            gdFINALMEM[gdDmxAddr]-- ;
        }
        break;

    case '=':
        if ((k = parseDeciNum(512)) > 512)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else setEqual(k);
        break;

    case '*':
        if ((k = parseDeciNum(1)) > 1)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else geArtnetOn = (uint8_t)k;    //1 = Artnet ON (default), 0=OFF
        break;

   case 'C':
        if ((k = parseDeciNum(512)) > 512)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else if (k < 24)  {                 //shortest allowed DMX cycle
            sendCmdB('?');
            gcPushByte = 0;
        }
        else gdDmxMax = k;
        break;

    case 'D':
            sendCmdB(0x0D);
            sendString(AgainStr);
            do  {
                recvword = pullCmdRxBuf();
            }   while (recvword == 0) ;
            if ((recvword == 0x8044) || (recvword == 0x8064)) { //'d' or 'D'
                for (k=1; k<513; k++)   {
                    gdDMXBUF[k] = 0 ;     
                    gdFINALMEM[k] = 0;    //Bugfix 27July20
                }
            }
            else  {
                sendCmdB('?');
                gcPushByte = 0;
            }
            break;

    case '@':
        if ((k = parseDeciNum(127)) > 127)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else popPresetPage(k);
        break;
   case '~':
        if ((k = parseDeciNum(127)) > 127)  {
            sendCmdB('?');
            gcPushByte = 0;
        }
        else pushPresetPage(k);
        break;
    case 0x0A:
        sendCmdB(0x0A);
        break;
    case 0x0D:
        sendCmdB(0x0D);
        break;
    case 0x20:
        break;      //ignore SPACE
    default:
        sendCmdB('?');
        gcPushByte = 0;
        break;
    }
}


void MidiCmd_Interpreter(uint32_t midiMsg)  // ******************************
// **************************************************************************
{
uint8_t index, channel, status, data1, data2;
uint16_t k;

    index = (uint8_t) (midiMsg & 0x000000FF);
//exclusively accept MIDI NOTE ON/OFF, CONTROL CHANGE or PROGRAM CHANGE
    if ((index!=0x08)&&(index!=0x09)&&(index!=0x0B)&&(index!=0x0C)) return;
    status = (uint8_t) ((midiMsg>>8) & 0x000000FF);
    data1 = (uint8_t) ((midiMsg>>16) & 0x000000FF);
    data2 = (uint8_t) ((midiMsg>>24) & 0x000000FF);

    if (index == 0x08)  {
        index = 0x09;
        status |= 0x10;     //convert NOTE OFF to NOTE ON
        data2 = 0;          //with zero velocity
    }
    channel = status & 0x0F;
    if (index == 0x09){
        if ((channel < gComParam.Channel) ||
            (channel > (gComParam.Channel + 3)))  return; //Bugfix 27July20
    }
    else if (channel != gComParam.Channel) return;

#ifdef DEBUG
    sendX8(index);
    sendX8(status);
    sendX8(data1);
    sendX8(data2);
#endif

    switch (index)  {
    case 0x09:
        if ((data2 == 0) && (gcIgnoreVelo0))  break;
        gdDmxAddr = data1 + 128 * (channel - gComParam.Channel);
        if (gdDmxAddr == 0) gdDmxAddr = 512;
        data2 = data2 << 1;                 //adjust level 0 ... 0xFE(254)
        if (data2 == 0xFE) data2 = 0xFF;
        if ((data2 == 2) && (gcIgnoreVelo0)) data2 = 0;  //special case !!
        setDmxLevel(data2);
        break;

    case 0x0B:
       if (data1 == 1) gdFadeTime = data2 * 5;
        else if ((data1 >= 0x50) && (data1 < 0x54))  {
            if ((data1 == 0x50) && (data2 == 0x00)) gdDmxAddr = 512;
            else gdDmxAddr = 128 * (data1 - 0x50) + data2;
        }
        else if (data1 == 0x54)  {
			data2 = data2 <<1;
			setDmxLevel(data2);     //Bugfix 27July20
        }
        else if (data1 == 0x55)  {
			data2 = (data2 <<1) + 1;
			setDmxLevel(data2);     //Bugfix 27July20
        }
        else if ((data1 == 0x56) && (gdDmxAddr < 512))  {
            gdDmxAddr++ ;
            data2 = data2 <<1;
            setDmxLevel(data2);
        }
        else if ((data1 == 0x57) && (gdDmxAddr < 512)) {
            gdDmxAddr++ ;
            data2 = (data2 <<1) + 1;
            setDmxLevel(data2);
        }
        else if ((data1 == 0x24) && (data2 != 0)) {
                sendAddrLevel();
                if (data2 == 1)  break;
                data2--;
                if (data2 > 9) data2 = 9;
            for (k=gdDmxAddr+1;k < (gdDmxAddr+1+data2);k++) {
                if (k < 513) sendLevel(k);
            }
        }
        else if (data1 == 0x40)  {
            if (data2 < 5) break;		//ignore bad input	bugfix 27July20
            else gdDmxMax = (data2+1) << 2;     //in steps of 4 DMX Tx slots
        }
        else if (data1 == 0x48)  setEqual(data2);
        else if (data1 == 0x60) popPresetPage(data2);
        else if (data1 == 0x70) pushPresetPage(data2);
        else if (data1 == 0x77)  {
            data1 = data2;
            goto PROGRAM_CHANGE;
        }
        break;
    case 0x0C:
PROGRAM_CHANGE:
        if ((data1 == 8) && (gdDMXBUF[gdDmxAddr] != 0x00)) {
            gdDMXBUF[gdDmxAddr]-- ;
            gdFINALMEM[gdDmxAddr]-- ;
        }
        else if ((data1 == 9) && (gdDMXBUF[gdDmxAddr] != 0xFF)) {
            gdDMXBUF[gdDmxAddr]++ ;
            gdFINALMEM[gdDmxAddr]++ ;
        }
        else if (data1 == 0x24) sendAddrLevel();
        else if (data1 == 0x40) geArtnetOn = 0;
        else if (data1 == 0x41) geArtnetOn = 1;
        else if (data1 == 0x78) gcIgnoreVelo0 = 0;
        else if (data1 == 0x79) gcIgnoreVelo0 = 0xFF;
        else if (data1 == 0x7F) {
            for (k=1;k<513;k++)  {
            gdDMXBUF[k] = 0 ; 
            gdFINALMEM[k] = 0;    //Bugfix 27July20
            }
        }
        break;
    default:
        break;
    }
}


void  Setup_Interpreter(uint8_t opcode)   // ******************************
// ***************************************************************************
{
uint32_t  parameter;
uint16_t recvword;
const uint8_t AgainStr[] = {"Save to Flash?->press again "};

    if ((opcode > 0x60) && (opcode <0x7B)) opcode -= 0x20;
    if ((opcode == 0x0D) || (opcode == 0x0A)) {
        sendCmdB(0x0D);
        return;
    };

    switch(opcode)  {


        case 'M':
            sendString(gcSetStr);
            sendString(gcMacIdStr);
			sendCmdB(':');
            sendCmdB(0x20);
            setMacID();
            sendCmdB(0x0D);
            break;

        case 'G':
            sendString(gcSetStr);
            sendString(gcGatewayStr);
            sendCmdB(':');
			sendCmdB(0x20);
            if (setEtherIP() == 0) {
                sendCmdB('?');
                break;
            }
            gComParam.Gateway = gcEtherIP;
			sendCmdB(0x0D);
            break;;

       case 'S':
            sendString(gcSetStr);
            sendString(gcSubnetStr);
            sendCmdB(':');
			sendCmdB(0x20);
            if (setEtherIP() == 0) {
                sendCmdB('?');
                break;
            }
            gComParam.Subnet = gcEtherIP;
			sendCmdB(0x0D);
            break;

        case 'I':
            sendString(gcSetStr);
            sendString(gcLocalIpStr);
            sendCmdB(':');
			sendCmdB(0x20);
            if (setEtherIP() == 0) {
                sendCmdB('?');
                break;
            }
            gComParam.LocalIP = gcEtherIP;
			sendCmdB(0x0D);
            break;

        case 'A':
            sendString(gcSetStr);
            sendString(gcArtIpStr);
            sendCmdB(':');
			sendCmdB(0x20);
            parameter = parseDeciNum(10);
            if ((parameter != 0) && (parameter != 2) && (parameter != 10))  {
                sendCmdB('?');
                break;
            }
            gComParam.ArtnetIP = parameter;
			sendCmdB(0x0D);
            break;

        case 'N':
            sendString(gcSetStr);
            sendString(gcArtSubnetStr);
            sendCmdB(':');
			sendCmdB(0x20);
            if ((parameter = parseDeciNum(15)) > 15)  {
                sendCmdB('?');
                break;
            }
            gComParam.ArtSubnet = parameter;
			sendCmdB(0x0D);
            break;

       case 'U':
            sendString(gcSetStr);
            sendString(gcArtUnivStr);
            sendCmdB(':');
			sendCmdB(0x20);
            if ((parameter = parseDeciNum(15)) > 15)  {
                sendCmdB('?');
                break;
            }
            gComParam.ArtUniverse = parameter;
			sendCmdB(0x0D);
            break;

        case 'C':
            sendString(gcSetStr);
            sendString(gcChannelStr);
            sendCmdB(':');
			sendCmdB(0x20);
			setMidiChannel();
			sendCmdB(0x0D);
             break;

        case 'V':
            sendString(gcSetStr);
         	sendString(gcVidPidStr);
			sendCmdB(':');
			sendCmdB(0x20);
            if (setVidPid() == 0) sendCmdB('?');
			sendCmdB(0x0D);
            break;

        case '!':                   //list actual configuration
            listSetup();
			sendCmdB(0x0D);
            break;

        case '?':
            sendCmdB(0x0D);
            listCommands();
            break;

        case '|':
            sendCmdB(0x0D);
            sendString(AgainStr);
            do  {
                recvword = pullCmdRxBuf();
            }  while (recvword == 0) ;
            if (recvword == 0x807C)  {      //'|'
                pushFlashPage15();
                sendCmdB(0x20);
                sendCmdB('o');
                sendCmdB('k');
                sendCmdB(0x0D);
                break;
            }

        case 0x0D:
            sendCmdB(0x0D);
            break;

        default: {
            sendCmdB('?');
            break;
        }
    }
}


uint32_t parseHexNibble(void)   //*******************************************
// **************************************************************************
{
uint16_t recvword;
uint32_t nibble;

    while ((recvword = pullCmdRxBuf()) == 0) ;
    nibble = (uint32_t)(recvword & 0x00FF);
    if ((nibble == 0x0A)||(nibble == 0x0D)||(nibble == '.')) return 0xFFFFFFFE;
    if  ((nibble >= 0x30) && (nibble <= 0x39))  {
        sendCmdB((uint8_t)nibble);
        return (nibble - 0x30);
    }
    if ((nibble > 0x60) && (nibble < 0x7B))  nibble -= 0x20; //force uppercase
    if ((nibble >= 'A') && (nibble <= 'F')) {
        sendCmdB((uint8_t)nibble);
        return (nibble - 0x37);
    }
    else return 0xFFFFFFFF;

}


uint32_t  parseHexByte(void)   // ********************************************
// ***************************************************************************
{
uint32_t nibble1, nibble2;

        if ((nibble1 = parseHexNibble())  == 0xFFFFFFFE) return 0xFFFFFFFF;
        else if (nibble1 > 0x0000000F) return 0xFFFFFFFF;
        if ((nibble2 = parseHexNibble()) == 0xFFFFFFFE) return nibble1;
        else if (nibble2 > 0x0000000F)  return 0xFFFFFFFF;
        if (parseHexNibble() != 0xFFFFFFFE) return 0xFFFFFFFF;
        return (uint32_t)((nibble1 << 4) | nibble2);
}


uint32_t parseDeciNum(uint32_t limit)   // **********************************
//unsigned, 0 ...max 65535
// **************************************************************************
{
uint32_t decinum;
uint16_t recvword;
uint8_t digit;

    if (limit > 65535) return 0xFFF10000;
    decinum = 0;
    while (decinum <= limit)  {
        while ((recvword = pullCmdRxBuf()) == 0) ;
        digit = (uint8_t)(recvword & 0x00FF);

        if ((digit < 0x30) || (digit > 0x39 )) {
            gcPushByte = digit;     //flexible delimiter, will be evaluated
                                    //as maybe next command code
            return decinum;
        }
        sendCmdB(digit);
        digit -= 0x30;              //ASCII>num
        decinum  = (decinum * 10) + digit;

//avoid necessity of final CRR when max number of places is reached
        if (decinum > limit) {
            gcPushByte = 0;
            return 0xFFFFFFFF;
        }
        if (limit < 10)  {
            gcPushByte = 0;
            return decinum;
        }
        if (decinum >= 10000) {
            gcPushByte = 0;
            return decinum;
        }
        else if ((decinum >= 1000) && (limit < 10000))  {
            gcPushByte = 0;
            return decinum;
        }
        else if ((decinum >= 100) && (limit < 1000))  {
            gcPushByte = 0;
            return decinum;
        }
        else if ((decinum >= 10) &&(limit < 100))  {
            gcPushByte = 0;
            return decinum;
        }
    }
    gcPushByte = 0;
    return decinum;             //exactly == limit
}


uint8_t parseDeciNibble(void)   // ******************************************
// **************************************************************************
{
uint32_t number;

    if ((number = parseDeciNum(15)) > 15) {
        sendCmdB('?');
        return 0xFF;
    }
    return (uint8_t) number;
}


void sendS3(uint32_t number)   // ************************************************
// **************************************************************************
{
    if (number < 100) sendCmdB('0');
    if (number < 10) sendCmdB('0');
    sendS32(number);
}


void sendS2(uint32_t number)   // ************************************************
// **************************************************************************
{
    if (number < 10) sendCmdB(0x20);
    sendS32(number);

}


uint8_t setVidPid(void)  // ****************************************************
// **************************************************************************
{
uint8_t nibble;
uint16_t Vid, Pid;
const uint8_t VidStr[] = {"Vid="};
const uint8_t PidStr[] = {"Pid="};
    sendString(VidStr);
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Vid = (uint16_t)nibble;
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Vid = (Vid << 4) + (uint16_t)nibble;                // *16
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Vid = (Vid << 4) + (uint16_t)nibble;
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Vid = (Vid << 4) + (uint16_t)nibble;
    sendCmdB(' ');

    sendString(PidStr);
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Pid = (uint16_t)nibble;
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Pid = (Pid << 4) + (uint16_t)nibble;
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Pid = (Pid << 4) + (uint16_t)nibble;
    if ((nibble = parseHexNibble()) > 0x000000FF) return 0;
    Pid = (Pid << 4) + (uint16_t)nibble;

//only write to Shadow after all items are entered correctly
    gVidShadow = Vid;
    gPidShadow = Pid;
    return 1;
}


uint8_t setMacID(void)  // *****************************************************
// **************************************************************************
{
uint32_t  macHi, macLo, temp;

    temp = parseHexByte();
    if (temp > 255) return 0;
    macHi = temp;
    sendCmdB('.');
    temp = parseHexByte();
    if (temp > 255) return 0;
    macHi = (macHi << 8) | temp;
    sendCmdB('.');
    temp = parseHexByte();
    if (temp > 255) return 0;
    macLo = temp;
    sendCmdB('.');
    temp = parseHexByte();
    if (temp > 255) return 0;
    macLo = (macLo << 8) | temp;
    sendCmdB('.');
    temp = parseHexByte();
    if (temp > 255) return 0;
    macLo = (macLo << 8) | temp;
    sendCmdB('.');
    temp = parseHexByte();
    if (temp > 255) return 0;
    macLo = (macLo << 8) | temp;
// replace not before MAC ID completely parsed
    gComParam.MacHi = macHi;
    gComParam.MacLo = macLo;
    return 1;
}


uint8_t setEtherIP(void)  // ************************************************
// **************************************************************************
{
uint32_t  etherIP, temp;

    temp = parseDeciNum(255);
    if (temp > 255) return 0;
    etherIP = temp;
    sendCmdB('.');
    temp = parseDeciNum(255);
    if (temp > 255) return 0;
    etherIP = (etherIP << 8) | temp;
    sendCmdB('.');
    temp = parseDeciNum(255);
    if (temp > 255) return 0;
    etherIP = (etherIP << 8) | temp;
    sendCmdB('.');
    temp = parseDeciNum(255);
    if (temp > 255) return 0;
    etherIP = (etherIP << 8) | temp;
// replace not before ID completely parsed
    gcEtherIP = etherIP;
    return 1;
}


void setMidiChannel(void)  // ***********************************************
//get the common MIDI channel in decimal format 1...16
// **************************************************************************
{
uint32_t number;

    if(((number = parseDeciNum(16)) > 16) || (number == 0)) {
        sendCmdB(' ');
        sendCmdB('1');
        sendCmdB('-');
        sendCmdB('1');
        sendCmdB('6');
        sendCmdB('!');
        sendCmdB('?');
        return;
    }
    gComParam.Channel = number - 1;
}


void listCommands(void)  // *************************************************
// **************************************************************************
{
const uint8_t SetupStr[] = {"List of Setup Commands:"};
const uint8_t ActualStr[] = {"list Actual Setup"};
const uint8_t FlashStr[] = {"save new setup permanently and activate it"};
const uint8_t TypeStr[] = "USB/DMX+Artnet Generator v" ;
const uint8_t CopyStr[] = " (c)2017 W.Schemmert";

    sendString(TypeStr);
    sendCmdB(HARDWARE_VERSION + 0x30);
    sendCmdB('.');
    sendCmdB(REVNUM + 0x30);
    sendString(CopyStr);
    sendCmdB(0x0D);
    sendString(SetupStr);
    sendCmdB(0x0D);

    sendCmdB('!');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(ActualStr);
    sendCmdB(0x0D);

    sendCmdB('V');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcSetStr);
    sendString(gcVidPidStr);
    sendCmdB(0x0D);

    sendCmdB('M');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcSetStr);
    sendString(gcMacIdStr);
    sendCmdB(0x0D);

    sendCmdB('G');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcSetStr);
    sendString(gcGatewayStr);
    sendCmdB(0x0D);

    sendCmdB('I');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcSetStr);
    sendString(gcLocalIpStr);
    sendCmdB(0x0D);

    sendCmdB('S');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcSetStr);
    sendString(gcSubnetStr);
    sendCmdB(0x0D);

    sendCmdB('A');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcSetStr);
    sendString(gcArtIpStr);
    sendCmdB(0x0D);

    sendCmdB('N');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcArtSubnetStr);
    sendString(gcSubnetStr);
    sendCmdB(0x0D);

    sendCmdB('U');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcArtUnivStr);
    sendString(gcSubnetStr);
    sendCmdB(0x0D);

    sendCmdB('C');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(gcSetStr);
    sendString(gcChannelStr);
    sendCmdB(0x0D);

    sendCmdB('|');
    sendCmdB(':');
    sendCmdB(0x20);
    sendString(FlashStr);
    sendCmdB(0x0D);
}


void listSetup(void)  // *****************************************************
// **************************************************************************
{
const uint8_t SetupStr[] = {"List of Actual Setup:"};

    sendCmdB(0x0D);
    sendString(SetupStr);
    sendCmdB(0x0D);

    sendX16(gVidShadow & 0x0000FFFF);
    sendCmdB('/');
    sendCmdB(0x20);
    sendX16(gPidShadow & 0x0000FFFF);
    sendString(gcVidPidStr);
    sendCmdB(0x0D);

    sendN8((gComParam.MacHi>>8) & 0xFF);
    sendCmdB('.');
    sendN8(gComParam.MacHi & 0xFF);
    sendCmdB('.');
    sendN8((gComParam.MacLo >> 24) & 0xFF);
    sendCmdB('.');
    sendN8((gComParam.MacLo >> 16) & 0xFF);
    sendCmdB('.');
    sendN8((gComParam.MacLo >> 8) & 0xFF);
    sendCmdB('.');
    sendN8(gComParam.MacLo & 0xFF);
    sendString(gcMacIdStr);
    sendCmdB(0x0D);

    sendS3((gComParam.Gateway >> 24) & 0xFF);
    sendCmdB('.');
    sendS3((gComParam.Gateway >> 16) & 0xFF);
    sendCmdB('.');
    sendS3((gComParam.Gateway >> 8) & 0xFF);
    sendCmdB('.');
    sendS3(gComParam.Gateway & 0xFF);
    sendString(gcGatewayStr);
    sendCmdB(0x0D);

    sendS3((gComParam.LocalIP >> 24) & 0xFF);
    sendCmdB('.');
    sendS3((gComParam.LocalIP >> 16) & 0xFF);
    sendCmdB('.');
    sendS3((gComParam.LocalIP >> 8) & 0xFF);
    sendCmdB('.');
    sendS3(gComParam.LocalIP & 0xFF);
    sendString(gcLocalIpStr);
    sendCmdB(0x0D);

    sendS3((gComParam.Subnet >> 24) &0xFF);
    sendCmdB('.');
    sendS3((gComParam.Subnet >> 16) & 0xFF);
    sendCmdB('.');
    sendS3((gComParam.Subnet >> 8) & 0xFF);
    sendCmdB('.');
    sendS3(gComParam.Subnet & 0xFF);
    sendString(gcSubnetStr);
    sendCmdB(0x0D);

    sendS2(gComParam.ArtnetIP & 0x0000000F);
    sendString(gcArtIpStr);
    sendCmdB(0x0D);

    sendS2(gComParam.ArtSubnet & 0x0000000F);
    sendString(gcArtSubnetStr);
    sendCmdB(0x0D);

    sendS2(gComParam.ArtUniverse & 0x0000000F);
    sendString(gcArtUnivStr);
    sendCmdB(0x0D);

    sendS2((gComParam.Channel & 0x0000000F) +1);
    sendString(gcChannelStr);
    sendCmdB(0x0D);
}

// ##########################################################################
